/////////////////////////////////////////////////////////////////////////////
//                                                                         //
// grid.cpp                                                                //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////
//                                                                         //
// Grid control library - Version 1.20 - 1 nov 1994                        //
// Original code: H. Picot                                                 //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

#include "griddll.h"

// Frame3d ------------------------------------------------------------------
void Frame3d (TDC &dc, TRect &rc, int w, TColor topleft, TColor bottomright)
   {
   Frame3d (dc, rc.left, rc.top, rc.right-1, rc.bottom-1, w,
            topleft, bottomright);
   }

void Frame3d (TDC &dc, int x0, int y0, int x1, int y1, int w,
              TColor topleft, TColor bottomright)
   {
   for (int i = 0; i < w; i++)
      {
      dc.SelectObject (TPen (topleft));
      dc.MoveTo (x1, y0);
      dc.LineTo (x0, y0);
      dc.LineTo (x0, y1);
      dc.RestorePen ();

      dc.SelectObject (TPen (bottomright));
      dc.LineTo (x1, y1);
      dc.LineTo (x1, y0);
      dc.RestorePen ();
      if (i < w-1)
         {
         x0++;
         y0++;
         x1--;
         y1--;
         }
      }
   }


// TGrid::Frame ---------------------------------------------------------- //
DEFINE_RESPONSE_TABLE1 (TGrid::Frame, TControl)
   EV_WM_ERASEBKGND,
END_RESPONSE_TABLE;

TGrid::Frame::Frame (TGrid *g) : TControl (g->Parent, -1, 0, 0, 0, 0, 0, 0)
   {
   grid = g;
   Attr.Style &= (~WS_TABSTOP);
   Attr.Style |= WS_BORDER|WS_CLIPSIBLINGS;
   }


void TGrid::Frame::Paint (TDC &dc, BOOL, TRect &clip)
   {
   TPen pen (GetSysColor (COLOR_WINDOWFRAME));
   int right = (grid->vscroll ? grid->vscroll->Attr.W-1 : 0);
   int bottom = (grid->hscroll ? grid->hscroll->Attr.H-1 : 0);
   int left = (grid->vbar ? grid->vbar->Attr.W-1 : 0);
   int top = (grid->title ? grid->title->Attr.H-1 : 0);

   // upper left corner
   if (grid->title && grid->vbar)
      {
      TRect rc (0, 0, left-1, top-1);
      if (rc.Touches (clip))
         DrawCorner (dc, rc);
      }

   // upper right corner
   if (grid->title && grid->vscroll)
      {
      TRect rc (Attr.W-right-2, 0, Attr.W-2, top-1);
      if (rc.Touches (clip))
         DrawCorner (dc, rc);
      }

   // lower left corner
   if (grid->hscroll && grid->vbar)
      {
      TRect rc (0, Attr.H-bottom-2, left-1, Attr.H-2);
      if (rc.Touches (clip))
         DrawCorner (dc, rc);
      }

   // lower right corner
   if (grid->hscroll && grid->vscroll)
      {
      TRect rc (Attr.W-right-2, Attr.H-bottom-2, Attr.W-2, Attr.H-2);
      if (rc.Touches (clip))
         DrawCorner (dc, rc);
      }
   }

void TGrid::Frame::DrawCorner (TDC &dc, TRect &rc)
   {
   Frame3d (dc, rc, 1,
            GetSysColor(COLOR_BTNHIGHLIGHT), GetSysColor (COLOR_BTNSHADOW));
   rc.Inflate (-1,-1);
   dc.FillRect (rc, TBrush (GetSysColor (COLOR_BTNFACE)));
   }


// TGrid::Element -------------------------------------------------------- //
TGrid::Element::Element (TControl *c, Element *prev, TGrid *g, const char *t)
   {
   next = 0;
   control = c;

   if (prev)
      prev->next = this;
   previous = prev;

   width = 0;
   left = 0;
   align = TGrid::Left;
   grid = g;
   title = t;
   }


TGrid::Element::~Element ()
   {
   if (next)
      {
      delete next;
      next = 0;
      }
   }


BOOL TGrid::Element::SetFocus ()
   {
   if (control && control->IsWindow () && control->IsWindowEnabled () &&
       (control->Attr.Style&WS_TABSTOP)==WS_TABSTOP && control->SetFocus ())
      {
      grid->current = this;
      return TRUE;
      }
   else
      return FALSE;
   }


// return the width of control owned by column --------------------------- //
int TGrid::Element::CtrlWidth () const
   {
   switch (type)
      {
      case gEdit:
         return dynamic_cast<TGEdit *>(control)->Width();
      case gStatic:
         return dynamic_cast<TGStatic *>(control)->Width();
      case gCheck:
         return dynamic_cast<TGCheckBox*>(control)->Width();
      case gCombo:
         return dynamic_cast<TGComboBox*>(control)->Width();
      case gButton:
         return dynamic_cast<TGButton*>(control)->Width();
      }
   return 0;
   }


// return the height of control owned by column -------------------------- //
int TGrid::Element::CtrlHeight () const
   {
   switch (type)
      {
      case gEdit:
         return dynamic_cast<TGEdit *>(control)->Height();
      case gStatic:
         return dynamic_cast<TGStatic *>(control)->Height();
      case gCheck:
         return dynamic_cast<TGCheckBox*>(control)->Height();
      case gCombo:
         return dynamic_cast<TGComboBox*>(control)->Height();
      case gButton:
         return dynamic_cast<TGButton*>(control)->Height();
      }
   return 0;
   }


void TGrid::Element::SetData (const char *data)
   {
   if (data == 0)
      data = "";
      
   switch (type)
      {
      case gStatic:
      case gEdit:
         {
         TStatic *ed = dynamic_cast<TStatic *>(control);
         if (ed)
            ed->SetText (data);
         break;
         }

      case gCombo:
         {
         TGComboBox *ed = dynamic_cast<TGComboBox *>(control);
         if (ed)
            ed->SetText (data);
         break;
         }

      case gCheck:
         {
         TGCheckBox *ed = dynamic_cast<TGCheckBox *> (control);
         if (ed)
            ed->SetCheck (*data != '0' ? BF_CHECKED : BF_UNCHECKED);
         break;
         }

      case gButton:
         {
         TGButton *ed = dynamic_cast<TGButton *> (control);
         if (ed)
            ed->SetCaption (data);
         break;
         }
      }
   }


void TGrid::Element::GetData (char * & data, int l) const
   {
   switch (type)
      {
      // case gButton: not used

      case gStatic:
      case gEdit:
         {
         TStatic *ed = dynamic_cast<TStatic *>(control);
         if (ed)
            ed->GetText (data, l);
         break;
         }

      case gCombo:
         {
         TGComboBox *ed = dynamic_cast<TGComboBox *>(control);
         if (ed)
            {
            int n = ed->GetSelIndex ();
            if (n >= 0)
               ed->GetString (data, n);
            else
               ed->GetText (data, l);
            }
         break;
         }

      case gCheck:
         {
         TGCheckBox *ed = dynamic_cast<TGCheckBox *>(control);
         if (ed)
            strcpy (data, ed->GetCheck() == BF_CHECKED ? "1" : "0");
         break;
         }
      }
   }


int TGrid::Element::GetDataLength () const
   {
   switch (type)
      {
      // case gButton: not used

      case gStatic:
      case gEdit:
         {
         TStatic *ed = dynamic_cast<TStatic *>(control);
         if (ed)
            return ed->GetTextLen () + 1;
         break;
         }

      case gCombo:
         {
         TGComboBox *ed = dynamic_cast<TGComboBox *>(control);
         if (ed)
            {
            int n = ed->GetSelIndex ();
            if (n >= 0)
               return ed->GetStringLen (n) + 1;
            else
               return ed->GetTextLen () + 1;
            }
         break;
         }

      case gCheck:
         {
         TGCheckBox *ed = dynamic_cast<TGCheckBox *>(control);
         if (ed)
            return 2;
         break;
         }
      }

   return 0;
   }


void TGrid::Element::SetTitle (const char *newtitle)
   {
   title = newtitle;
   if (grid->title && left < grid->Attr.W && left+width >= 0)
      {
      TRect rc (left+1, 1, left+width-2, grid->title->Attr.H-2);
      grid->title->InvalidateRect (rc, FALSE);
      }
   }


// VScroll --------------------------------------------------------------- //
TGrid::VScroll::VScroll (TGrid *g)
     : TScrollBar (g->Parent, -1, 0, 0, 0, 0, FALSE)
   {
   grid = g;
   Attr.Style &= (~WS_TABSTOP);
   }


void TGrid::VScroll::SBBottom ()
   {
   grid->Go (grid->nLines-1);
   }


void TGrid::VScroll::SBLineDown ()
   {
   grid->Move (1);
   }


void TGrid::VScroll::SBLineUp ()
   {
   grid->Move (-1);
   }


void TGrid::VScroll::SBPageDown ()
   {
   grid->Move (grid->visibleLines);
   }


void TGrid::VScroll::SBPageUp ()
   {
   grid->Move (-grid->visibleLines);
   }


void TGrid::VScroll::SBThumbPosition (int pos)
   {
   grid->Go (pos);
   }


void TGrid::VScroll::SBThumbTrack (int pos)
   {
   grid->Go (pos);
   }
               

void TGrid::VScroll::SBTop ()
   {
   grid->Go (0);
   }


// HScroll --------------------------------------------------------------- //
TGrid::HScroll::HScroll (TGrid *g)
     : TScrollBar (g->Parent, -1, 0, 0, 0, 0, TRUE)
   {
   grid = g;
   Attr.Style &= (~WS_TABSTOP);
   }


void TGrid::HScroll::SBBottom ()
   {
   int min, max;
   GetRange (min, max);
   grid->SetHPos (max);
   }


void TGrid::HScroll::SBLineDown ()
   {
   grid->MoveHPos (LineMagnitude);
   }


void TGrid::HScroll::SBLineUp ()
   {
   grid->MoveHPos (-LineMagnitude);
   }


void TGrid::HScroll::SBPageDown ()
   {
   grid->MoveHPos (PageMagnitude);
   }


void TGrid::HScroll::SBPageUp ()
   {
   grid->MoveHPos (-PageMagnitude);
   }


void TGrid::HScroll::SBThumbPosition (int pos)
   {
   grid->SetHPos (pos);
   }


void TGrid::HScroll::SBThumbTrack (int pos)
   {
   grid->SetHPos (pos);
   }


void TGrid::HScroll::SBTop ()
   {
   grid->SetHPos (0);
   }


// TGrid::Title ---------------------------------------------------------- //
TGrid::Title::Title (TGrid *g)
     : TControl (g->Parent, -1, 0, 0, 0, 0, 0, 0)
   {
   grid = g;
   Attr.Style |= WS_BORDER;
   Attr.Style &= (~WS_TABSTOP);
   }


void TGrid::Title::SetupWindow ()
   {
   TControl::SetupWindow ();
   int h = CalcHeight ();
   int x = grid->Attr.X;
   int y = grid->Attr.Y-h;
   int w = grid->Attr.W;

   SetWindowPos (HWND_BOTTOM, x, y, w, h, SWP_NOACTIVATE);
   }


int TGrid::Title::CalcHeight ()
   {
   int h = 18;  // default
   if (IsWindow ())
      {
      TFont *f = grid->titleFont;
      if (f == 0)
         f = grid->gridFont;
      if (f)
         {
         TClientDC dc(*this);
         dc.SelectObject (*f);
         TEXTMETRIC tm;
         dc.GetTextMetrics (tm);
         dc.RestoreFont ();
         h = tm.tmHeight + 6;  // minimum
         }
      if (h < 10)
         h = 10;
      }
   return h;
   }


void TGrid::Title::Paint (TDC &dc, BOOL, TRect &clip)
   {
   dc.FillRect (clip, TBrush (GetSysColor (COLOR_BTNFACE)));
   dc.SelectClipRgn (TRegion (clip));
   TFont *font = grid->titleFont;
   if (font == 0)
      font = grid->gridFont;
   dc.SelectObject (*font);
   TPen pen (GetSysColor (COLOR_WINDOWFRAME));

   for (TGrid::Element *ctrl = grid->firstControl; ctrl; ctrl = ctrl->next)
      {
      if (ctrl->previous)
         {
         dc.SelectObject (pen);
         dc.MoveTo (ctrl->left-1, 0);
         dc.LineTo (ctrl->left-1, Attr.H);
         dc.RestorePen ();
         }

      DrawCell (dc, ctrl->left, ctrl->width, ctrl->title.c_str(), ctrl->align);
      }

   dc.RestoreFont ();
   }


void TGrid::Title::DrawCell (TDC &dc, int x, int w, const char *p,
                             TGrid::Align a)
   {
   w = x+w-2;
   int h = Attr.H-3;
   Frame3d (dc, x, 0, w, h, 1,
            GetSysColor(COLOR_BTNHIGHLIGHT), GetSysColor (COLOR_BTNSHADOW));

   if (p)
      {
      TRect rc (x+1, 1, w-1, h-1);
      dc.SetTextColor (GetSysColor (COLOR_BTNTEXT));
      dc.SetBkColor (GetSysColor (COLOR_BTNFACE));
      dc.DrawText (p, -1, rc, (WORD)(a|DT_SINGLELINE|DT_VCENTER));
      }
   }


// TGrid::VerticalBar   -------------------------------------------------- //
DEFINE_RESPONSE_TABLE1 (TGrid::VerticalBar, TControl)
   EV_WM_LBUTTONDOWN,
END_RESPONSE_TABLE;

TGrid::VerticalBar::VerticalBar (TGrid *g)
     : TControl (g->Parent, -1, 0, 0, 0, 0, 0, 0)
   {
   grid = g;
   lineNumber = TRUE;
   charWidth = 4;
   Attr.Style |= WS_BORDER;
   Attr.Style &= (~WS_TABSTOP);
   }


void TGrid::VerticalBar::SetupWindow ()
   {
   TControl::SetupWindow ();
   int w = CalcWidth ();
   int h = grid->Attr.H;
   int x = grid->Attr.X - w+1;
   int y = grid->Attr.Y;

   SetWindowPos (HWND_BOTTOM, x, y, w, h, SWP_NOACTIVATE);
   }


int TGrid::VerticalBar::CalcWidth ()
   {
   int w = 16;
   if (IsWindow ())
      {
      TFont *f = grid->titleFont;
      if (f == 0)
         f = grid->gridFont;
      if (f)
         {
         TClientDC dc(*this);
         dc.SelectObject (*f);
         TEXTMETRIC tm;
         dc.GetTextMetrics (tm);
         dc.RestoreFont ();
         w = tm.tmAveCharWidth * charWidth + 8;
         }
      if (w < 16)
         w = 16;   // default
      }
   return w;
   }


void TGrid::VerticalBar::SetUserText (int cwidth)
   {
   if (!IsWindow ())
      {
      lineNumber = FALSE;
      if (cwidth == 0)
         cwidth = 4;
      charWidth = cwidth;
      }
   }


void TGrid::VerticalBar::Paint (TDC &dc, BOOL, TRect &clip)
   {
   dc.SelectClipRgn (TRegion (clip));
   dc.FillRect (clip, TBrush (GetSysColor (COLOR_BTNFACE)));

   int lh = grid->lineHeight;
   int x0 = 0;
   int x1 = Attr.W-1;
   int y0 = 0;
   int y1 = lh-1;
   TPen pen (GetSysColor (COLOR_WINDOWFRAME));
   for (int i = 0; i < grid->visibleLines; i++, y0 += lh, y1 += lh)
      {
      TRect rc (x0, y0, x1, y1);
      if (rc.Touches (clip))
         {
         DrawCell (dc, rc, i);
         dc.SelectObject (pen);
         dc.MoveTo (0, rc.bottom);
         dc.LineTo (Attr.W, rc.bottom);
         dc.RestorePen ();
         }
      }
   }


char *TGrid::VerticalBar::Info (int line) const
   {
   static char string[10];
   char *info = 0;
   if (lineNumber)
      {
      sprintf (string, "%d", grid->firstLine+line+1);
      info = string;
      }
   else
      info = grid->TransferItem (grid->firstLine+line, 0);

   return info;
   }


void TGrid::VerticalBar::DrawCell (TDC &dc, TRect &rc, int line)
   {
   TRect clip = dc.GetClipBox ();
   dc.SelectClipRgn (TRegion (rc));
   Frame3d (dc, rc, 1,
            GetSysColor (COLOR_BTNHIGHLIGHT), GetSysColor (COLOR_BTNSHADOW));
   if (line == grid->activeLine)
      {
      if (grid->lineMarker != TGrid::ArrowOnly)
         ShowText (line);
      ShowMarker (line);
      }
   else
      ShowText (line);
   dc.SelectClipRgn (TRegion (clip));
   }


void TGrid::VerticalBar::ShowText (int line) const
   {
   TClientDC dc (HWindow);
   TRect rc = FillRect (dc, line, TRUE);
   if (grid->lineMarker == TGrid::ReverseColor)
      {
      TRect rc1 (rc);
      rc1.Inflate (2, 2);
      Frame3d (dc, rc1, 1, GetSysColor (COLOR_BTNHIGHLIGHT),
                           GetSysColor (COLOR_BTNSHADOW));
      }

   char *info = Info (line);
   if (info)
      {
      TFont *font = grid->titleFont;
      if (font == 0)
         font = grid->gridFont;
      dc.SelectObject (*font);
      dc.SetTextColor (GetSysColor (COLOR_BTNTEXT));
      dc.SetBkColor (GetSysColor (COLOR_BTNFACE));
      dc.DrawText (info, -1, rc, (WORD)(DT_RIGHT|DT_VCENTER|DT_SINGLELINE));
      dc.RestoreFont ();
      }
   }


void TGrid::VerticalBar::ShowMarker (int line) const
   {
   TClientDC dc (HWindow);
   TRect rc = FillRect (dc, line, FALSE);
   if (grid->lineMarker == TGrid::ReverseColor)
      {
      TRect rc1 (rc);
      rc1.Inflate (2, 2);
      dc.InvertRect (rc1);
      }
   else
      {
      dc.SelectObject (TPen (GetSysColor (COLOR_BTNTEXT)));
      int x = rc.left+4;
      int y0 = rc.top+2;
      int y1 = rc.bottom-2;
      while (y0 < y1 && x < rc.right)
         {
         dc.MoveTo (x, y0++);
         dc.LineTo (x++, y1--);
         }
      dc.RestorePen ();
      }
   }


TRect TGrid::VerticalBar::FillRect (TDC &dc, int line, BOOL erase) const
   {
   int lh = grid->lineHeight;
   int y = line*lh;
   TRect rc (1, y+1, Attr.W-3, y+lh-2);
   if (erase || grid->lineMarker == TGrid::ArrowOnly)
      dc.FillRect (rc, TBrush(GetSysColor (COLOR_BTNFACE)));
   return rc.InflatedBy(-1,-1);
   }

   
void TGrid::VerticalBar::EvLButtonDown (UINT modKeys, TPoint &point)
   {
   if (grid->lineHeight > 0)
      {
      // calc cell where mouse was clicked
      int line = (point.y / grid->lineHeight) + grid->firstLine;
      grid->Go (line);
      }
      
   TControl::EvLButtonDown (modKeys, point);
   }

// TGrid ----------------------------------------------------------------- //
DEFINE_RESPONSE_TABLE1 (TGrid, TControl)
   EV_WM_CTLCOLOR,
   EV_WM_ERASEBKGND,
   EV_WM_LBUTTONDBLCLK,
   EV_WM_LBUTTONDOWN,
   EV_WM_PAINT,
   EV_WM_SETCURSOR,
   EV_WM_SETFOCUS,
   EV_WM_WINDOWPOSCHANGING,
END_RESPONSE_TABLE;

TGrid::TGrid (TWindow *p, int id, int x0, int y0, int w0, int h0,
              TGrid::EnterKey ek, TGrid::TabKey tk, TGrid::LineMarker lm,
              int max, int atrib)
    : TControl  (p, id, "", x0, y0, w0, h0)
   {
   Init (max, ek, tk, lm, atrib);
   }


TGrid::TGrid (TWindow *p, int id, TGrid::EnterKey ek, TGrid::TabKey tk,
              TGrid::LineMarker lm, int max, int attrib)
    : TControl  (p, id)
   {
   Init (max, ek, tk, lm, attrib);
   }


void TGrid::Init (int max, TGrid::EnterKey ek, TGrid::TabKey tk, TGrid::LineMarker lm, int attrib)
   {
   Attr.Style = BS_OWNERDRAW|WS_CHILD|WS_VISIBLE|
                WS_CLIPSIBLINGS|WS_BORDER|WS_TABSTOP;
   firstControl = lastControl = current = 0;
   lineHeight = firstLine = activeLine = visibleLines = 0;
   nControls = hpos = totalWidth = 0;
   if (max >= 0)
      nLines = max;
   else
      nLines = 0;
   maxLines = max;
   lineColor = TColor::Gray;
   firstPaint = TRUE;
   lineSwapping = TRUE;
   enterKey = ek;
   tabKey = tk;
   lineMarker = lm;

   vscroll = (attrib & gVScroll) ? new VScroll (this) : 0;
   hscroll = (attrib & gHScroll) ? new HScroll (this) : 0;
   title   = (attrib & gTitle)   ? new Title   (this) : 0;
   vbar    = (attrib & gVBar)    ? new VerticalBar (this) : 0;

   frame = new Frame (this);

   gridFont = 0;
   titleFont = 0;

   ComboBmp       =
   CheckBoxesBmp  =
   BwccCheckBmp   =
   BwccUncheckBmp = 0;

   titleLogFont = 0;
   gridLogFont = new LOGFONT;
   memset (gridLogFont, 0, sizeof(LOGFONT));
   strcpy (gridLogFont->lfFaceName, "MS Sans Serif");
   gridLogFont->lfHeight = MulDiv(TScreenDC().GetDeviceCaps (LOGPIXELSY),8,72);
   gridLogFont->lfWeight = FW_BOLD;
   }


TGrid::~TGrid ()
   {
   if (gridLogFont)
      delete gridLogFont;
   if (titleLogFont)
      delete titleLogFont;
   if (firstControl)
      delete firstControl;
   if (gridFont)
      delete gridFont;
   if (titleFont)
      delete titleFont;
   if (ComboBmp)
      delete ComboBmp;
   if (CheckBoxesBmp)
      delete CheckBoxesBmp;
   if (BwccCheckBmp)
      delete BwccCheckBmp;
   if (BwccUncheckBmp)
      delete BwccUncheckBmp;
   }


TGrid::Element * TGrid::FindElement (TPoint &p) const
   {
   for (Element *ctrl = firstControl; ctrl; ctrl = ctrl->next)
      {
      if (TRect (ctrl->left, 0, ctrl->left+ctrl->width, Attr.H).Contains (p))
         return ctrl;
      }
   return 0;
   }


TGrid::Element *TGrid::FindElement (TControl *c) const
   {
   for (Element *ctrl = firstControl; ctrl; ctrl = ctrl->next)
      if (ctrl->control == c)
         return ctrl;

   return 0;
   }


TGrid::Element *TGrid::FindElement (HWND hwnd) const
   {
   for (Element *ctrl = firstControl; ctrl; ctrl = ctrl->next)
      if (ctrl->control->HWindow == hwnd)
         return ctrl;

   return 0;
   }


TGrid::Element *TGrid::FindElement (int id) const
   {
   for (Element *ctrl = firstControl; ctrl; ctrl = ctrl->next)
      if (ctrl->control->GetId() == id)
         return ctrl;

   return 0;
   }


void TGrid::SetLineHeight (int h)
   {
   if (h == 0)
      return;
   visibleLines = Attr.H / (lineHeight=h+1);
   }


void TGrid::ResetPosition ()
   {
   Go (0);
   GotoFirstColumn ();
   }


void TGrid::Add (TControl *p, const char *t, Align a)
   {
   if (p)
      {
      Element *ge = new Element (p, lastControl, this, t);
      lastControl = ge;
      if (firstControl == 0)
         firstControl = ge;

      if (dynamic_cast<TGEdit *>(p))
         {
         ((TGEdit *)p)->grid = this;
         ge->type = gEdit;
         }
      else if (dynamic_cast<TGCheckBox *>(p))
         {
         ((TGCheckBox *)p)->grid = this;
         ge->type = gCheck;
         }
      else if (dynamic_cast<TGButton *>(p))
         {
         ((TGButton *)p)->grid = this;
         ge->type = gButton;
         }
      else if (dynamic_cast<TGComboBox *>(p))
         {
         ((TGComboBox *)p)->grid = this;
         ge->type = gCombo;
         }
      else if (dynamic_cast<TGStatic *>(p))
         ge->type = gStatic;

      nControls++;

      if (ge->type == gButton)
         ge->align = Center;
      else
         ge->align = a;
      }
   }

   
void TGrid::Move (int delta)
   {
   Go (GetCurrentLine()+delta);
   }


void TGrid::Go (int pos)
   {
   // verifying pos
   if (pos >= nLines)
      {
      if (maxLines > 0)
         pos = nLines-1;
      else
         {
         // Ask permission to append;
         // AppendLine may change pos (passed by reference)
         if (!AppendNewLine (pos))
            return;

         SetNLines (pos+1);
         }
      }
      
   if (pos < 0)
      pos = 0;

   // if no change, abort function
   int oldCurrent = GetCurrentLine();
   if (pos == oldCurrent)
      return;

   // transfer data to memory
   TransferLine (oldCurrent, tdGetData);

   // check data
   if (!IsValidLine (oldCurrent))
      return;

   // set focus to frame, while changing current line
   if (frame->SetFocus () == NULL)
      return;   // WM_KILLFOCUS failed

   // remove markers
   if (vbar)
      vbar->ShowText (activeLine);

   // compute new position
   int delta = pos - oldCurrent;
   int oldFirstLine = firstLine;
   if (delta + activeLine >= 0 && delta + activeLine < visibleLines)
      activeLine += delta;
   else
      {
      firstLine += delta;
      if (firstLine+visibleLines > nLines && maxLines > 0)
         firstLine = nLines-visibleLines;
      if (firstLine < 0)
         firstLine = 0;
      int old = activeLine;
      activeLine = pos - firstLine;
      if (activeLine < 0)
         activeLine = 0;
      else if (activeLine >= visibleLines)
         activeLine = visibleLines-1;
      delta = activeLine-old;
      }

   // move controls if necessary
   if (delta)
      {
      HDWP hdwp = BeginDeferWindowPos (nControls);
      for (Element *ctrl = firstControl; ctrl; ctrl = ctrl->next)
         {
         TControl *c = ctrl->control;
         TWindowAttr *a = &c->Attr;
         hdwp = DeferWindowPos (hdwp, c->HWindow, 0, a->X,
                                a->Y+lineHeight*delta, 0, 0,
                                SWP_NOZORDER|SWP_NOSIZE|SWP_NOREDRAW);
         }
      EndDeferWindowPos (hdwp);
      }

   // redraw grid or scroll
   if (firstLine != oldFirstLine)
      {
      int scroll = oldFirstLine-firstLine;
      TRect vbclip (0, 0, vbar?vbar->Attr.W-2:0, vbar?vbar->Attr.H-2:0);
      TRect clip (0,0,Attr.W-2,Attr.H-2);
      if (scroll >= visibleLines || scroll <= -visibleLines)
         {
         InvalidateRect (clip, FALSE);
         if (vbar)
            vbar->InvalidateRect (vbclip, FALSE);
         }
      else
         {
         int s = scroll*lineHeight;
         ScrollWindowEx (0, s, 0, &clip, 0,0, SW_INVALIDATE);
         if (vbar)
            vbar->ScrollWindowEx (0, s, 0, &vbclip, 0,0, SW_INVALIDATE);
         }
      if (vbar)
         vbar->UpdateWindow ();
      }

   // transfer data from memory
   int newCurrent = GetCurrentLine ();
   TransferLine (newCurrent, tdSetData);

   // update scroller
   if (vscroll)
      vscroll->SetPosition (newCurrent);

   // update separators
   if (vbar && firstLine == oldFirstLine)
      vbar->ShowMarker (activeLine);

   // reset focus
   if (current)
      current->SetFocus ();
   }


BOOL TGrid::KeyDown (UINT key, int repeat)
   {
   BOOL ctrl  = (GetKeyState (VK_CONTROL) & 0x8000) ? TRUE : FALSE;
   BOOL shift = (GetKeyState (VK_SHIFT)   & 0x8000) ? TRUE : FALSE;
                           
   switch (key)
      {
      case VK_ESCAPE:   return VkEscape (ctrl, shift, repeat);
      case VK_INSERT:   return VkInsert (ctrl, shift, repeat);
      case VK_DELETE:   return VkDelete (ctrl, shift, repeat);
      case VK_PRIOR:    return VkPrior  (ctrl, shift, repeat);
      case VK_NEXT:     return VkNext   (ctrl, shift, repeat);
      case VK_UP:       return VkUp     (ctrl, shift, repeat);
      case VK_DOWN:     return VkDown   (ctrl, shift, repeat);
      case VK_RIGHT:    return VkRight  (ctrl, shift, repeat);
      case VK_LEFT:     return VkLeft   (ctrl, shift, repeat);
      case VK_RETURN:   return VkReturn (ctrl, shift, repeat);
      case VK_TAB:      return VkTab    (ctrl, shift, repeat);
      case VK_HOME:     return VkHome   (ctrl, shift, repeat);
      case VK_END:      return VkEnd    (ctrl, shift, repeat);
      }

   return FALSE;
   }


BOOL TGrid::VkEscape (BOOL ctrl, BOOL shift, int)
   {
   if (ctrl == FALSE && shift == FALSE)
      {
      Parent->CloseWindow (IDCANCEL);
      return TRUE;
      }
      
   return FALSE;
   }


BOOL TGrid::VkEnd (BOOL ctrl, BOOL, int)
   {
   if (ctrl)
      {
      GotoLastColumn ();
      return TRUE;
      }

   return FALSE;
   }


BOOL TGrid::VkHome (BOOL ctrl, BOOL, int)
   {
   if (ctrl)
      {
      GotoFirstColumn ();
      return TRUE;
      }
      
   return FALSE;
   }

   
BOOL TGrid::VkRight (BOOL, BOOL, int repeat)
   {
   for (int i = 0; i < repeat; i++)
      if (!GotoNextColumn ())
         break;
   return TRUE;
   }


BOOL TGrid::VkLeft (BOOL, BOOL, int repeat)
   {
   for (int i = 0; i < repeat; i++)
      if (!GotoPreviousColumn ())
         break;
   return TRUE;
   }


BOOL TGrid::VkUp (BOOL ctrl, BOOL, int repeat)
   {
   if (ctrl)
      {
      if (lineSwapping == FALSE)
         return FALSE;
      for (int i = 0; i < repeat; i++)
         {
         int l = firstLine+activeLine;
         if (l <= 0)
            break;
         SwapLines (l-1, l);
         Move (-1);
         }
      }
   else
      Move (-repeat);
   return TRUE;
   }


BOOL TGrid::VkDown (BOOL ctrl, BOOL, int repeat)
   {
   if (ctrl)
      {
      if (lineSwapping == FALSE)
         return FALSE;
      for (int i = 0; i < repeat; i++)
         {
         int l = firstLine+activeLine;
         if (l >= maxLines-1)
            break;
         SwapLines (l, l+1);
         Move (1);
         }
      }
   else
      Move (repeat);
   return TRUE;
   }


BOOL TGrid::VkNext (BOOL ctrl, BOOL, int repeat)
   {
   if (ctrl)
      Go (nLines-1);
   else
      Move (visibleLines*repeat);
   return TRUE;
   }


BOOL TGrid::VkPrior (BOOL ctrl, BOOL, int repeat)
   {
   if (ctrl)
      Go (0);
   else
      Move (-visibleLines*repeat);
   return TRUE;
   }


BOOL TGrid::VkReturn (BOOL ctrl, BOOL, int repeat)
   {
   if (ctrl)
      return FALSE;

   switch (enterKey)
      {
      case NextLine:
         {
         Move (repeat);
         GotoFirstColumn ();
         return TRUE;
         }

      case DoNothing:
         return TRUE;

      case EndDialog:
         return FALSE;

      case EnterNextColumn:
         {
         for (int i = 0; i < repeat; i++)
            {
            if (!GotoNextColumn ())
               {
               GotoFirstColumn ();
               Move (1);
               }
            }
         return TRUE;
         }
      }
   return FALSE;
   }


BOOL TGrid::VkTab (BOOL ctrl, BOOL shift, int repeat)
   {
   if (ctrl)
      {
      if (IsValidLine(activeLine+firstLine))
         {
         HWND hwnd = Parent->GetNextDlgTabItem (HWindow, shift);
         if (hwnd)
            ::SetFocus (hwnd);
         }
      return TRUE;
      }
   else if (tabKey == NextControl)
      return FALSE;
   else if (tabKey == AutoNextLine)
      {
      if (shift)
         {
         for (int i = 0; i < repeat; i++)
            {
            if (!GotoPreviousColumn ())
               {
               GotoLastColumn ();
               Move (-1);
               }
            }
         }
      else
         {
         for (int i = 0; i < repeat; i++)
            {
            if (!GotoNextColumn ())
               {
               GotoFirstColumn ();
               Move (1);
               }
            }
         }
      return TRUE;
      }
   else
      {
      if (shift)
         return VkLeft (ctrl, shift, repeat);
      else
         return VkRight (ctrl, shift, repeat);
      }
   }


BOOL TGrid::VkDelete (BOOL ctrl, BOOL, int)
   {
   if (ctrl)
      {
      if (DeleteLine (activeLine+firstLine))
         {
         TransferLine (activeLine+firstLine, tdSetData);
         frame->SetFocus ();
         Invalidate ();
         if (current)
            current->SetFocus ();
         }
      return TRUE;
      }
   return FALSE;
   }


BOOL TGrid::VkInsert (BOOL ctrl, BOOL, int repeat)
   {
   if (ctrl)
      {
      if (!IsValidLine (activeLine+firstLine))
         return TRUE;
      for (int i = 0; i < repeat; i++)
         {
         if (!InsertLine (activeLine+firstLine))
            return TRUE;
         TransferLine (activeLine+firstLine, tdSetData);
         }
      frame->SetFocus ();
      Invalidate ();
      if (current)
         current->SetFocus ();
      return TRUE;
      }
   return FALSE;
   }


BOOL TGrid::GotoPreviousColumn ()
   {
   Element *e = current;
   if (e == 0)
      e = lastControl;
   if (e)
      e = e->previous;
   for (; e; e = e->previous)
      {
      if (e->SetFocus ())
         return TRUE;
      }
   return FALSE;
   }


BOOL TGrid::GotoNextColumn ()
   {
   Element *e = current;
   if (e == 0)
      e = firstControl;
   if (e)
      e = e->next;
   for (; e; e = e->next)
      {
      if (e->SetFocus ())
         return TRUE;
      }
   return FALSE;
   }


void TGrid::GotoLastColumn ()
   {
   current = lastControl;
   if (lastControl && !lastControl->SetFocus ())
      GotoPreviousColumn ();
   }


void TGrid::GotoFirstColumn ()
   {
   current = firstControl;
   if (firstControl && !firstControl->SetFocus ())
      GotoNextColumn ();
   }


HBRUSH TGrid::EvCtlColor(HDC hDC, HWND hWndChild, UINT)
   {
   TWindow *wptr = GetWindowPtr (hWndChild);

   if (dynamic_cast<TGEdit *>(wptr))
      return dynamic_cast<TGEdit *>(wptr)->CtlColors (hDC);
   if (dynamic_cast<TGComboBox *>(wptr))
      return dynamic_cast<TGComboBox *>(wptr)->CtlColors (hDC);
   if (dynamic_cast<TGStatic *>(wptr))
      return dynamic_cast<TGStatic *>(wptr)->CtlColors (hDC);
   if (dynamic_cast<TGCheckBox *>(wptr))
      return dynamic_cast<TGCheckBox *>(wptr)->CtlColors (hDC);

   return 0;
   }


void TGrid::EvSetFocus (HWND)
   {
   GotoFirstColumn ();
   }


void TGrid::EvPaint ()
   {
   TPaintDC dc(*this);
   Paint(dc, dc.Ps.fErase, *(TRect*)&dc.Ps.rcPaint);
   }


void TGrid::Paint (TDC &dc, BOOL, TRect &clip)
   {
   // terminate initialization
   if (firstPaint)
      {
      firstPaint = FALSE;
      CalcPos ();
      Invalidate ();    // added
      return;
      }

   // clip drawing
   dc.SelectClipRgn (TRegion(clip));

   // Select font
   dc.SelectObject (*gridFont);

   // draw lines
   int y = 0;
   for (int i = 0; i < visibleLines; i++, y += lineHeight)
      {
      int pos = firstLine+i;
      if (maxLines < 0 && pos >= nLines)
         {
         if (!AppendNewLine (pos))
            {
            TRect rc (0, y, Attr.W, Attr.H);
            dc.FillRect (rc, TBrush (GetSysColor (COLOR_WINDOW)));
            break;
            }
         SetNLines (pos+1);
         }

      TRect current (0, y, Attr.W-1, y+lineHeight-1);
      if (!current.Touches (clip))
         continue;

      // text
      for (Element *ctrl = firstControl; ctrl; ctrl = ctrl->next)
         {
         // do not draw elements scrolled out horizontally
         if (ctrl->left + ctrl->width <= 0)
            continue;
         if (ctrl->left >= Attr.W)
            continue;

         TControl *c = ctrl->control;
         char *data = TransferItem (i+firstLine, c->GetId ());

         // drawing area
         TRect rc (ctrl->left, y, ctrl->left+ctrl->width-1, y+lineHeight-1);
         if (rc.Touches (clip))
            {
            switch (ctrl->type)
               {
               case gEdit:
                  dynamic_cast<TGEdit *>(c)->DrawCell
                                                 (dc, rc, data, ctrl->align);
                  break;
               case gStatic:
                  dynamic_cast<TGStatic *>(c)->DrawCell
                                                 (dc, rc, data, ctrl->align);
                  break;
               case gCombo:
                  dynamic_cast<TGComboBox *>(c)->DrawCell (dc, rc, data);
                  break;
               case gCheck:
                  dynamic_cast<TGCheckBox *>(c)->DrawCell
                                                 (dc, rc, data, ctrl->align);
                  break;
               case gButton:
                  dynamic_cast<TGButton *>(c)->DrawCell (dc, rc, data);
                  break;
               }
            }
         }
      }
   dc.RestoreFont ();

   // draw grid
   dc.SelectObject (TPen (lineColor));
   y = lineHeight-1;
   for (i = 1; i < visibleLines; i++, y += lineHeight)
      {
      dc.MoveTo (0, y);
      dc.LineTo (Attr.W-2, y);
      }

   for (Element *ctrl = firstControl->next; ctrl; ctrl = ctrl->next)
      {
      int x = ctrl->left-1;
      if (x >= 0 && x < Attr.W)
         {
         dc.MoveTo (x, 0);
         dc.LineTo (x, Attr.H-2);
         }
      }
   dc.RestorePen ();
   }


void TGrid::EvWindowPosChanging (WINDOWPOS far &wpos)
   {
   ChangeWindowPos (wpos.x, wpos.y, wpos.cx, wpos.cy, wpos.flags);
   TControl::EvWindowPosChanging (wpos);
   }


void TGrid::ChangeWindowPos (int &x, int &y, int &w, int &h, UINT flags)
   {
   // do nothing if initialization is not finished
   if (firstPaint)
      return;

   BOOL Move = ((flags & SWP_NOMOVE) == 0 ? TRUE : FALSE);
   BOOL Size = ((flags & SWP_NOSIZE) == 0 ? TRUE : FALSE);

   if (flags & SWP_NOMOVE)
      {
      x = Attr.X;
      y = Attr.Y;
      }

   if (vbar)
      {
      vbar->Attr.W = vbar->CalcWidth ();
      x += (vbar->Attr.W -1);
      w -= (vbar->Attr.W -1);
      }

   if (title)
      {
      title->Attr.H = title->CalcHeight ();
      y += (title->Attr.H -1);
      h -= (title->Attr.H -1);
      }

   if (vscroll)
      w -= (GetSystemMetrics(SM_CXVSCROLL)-1);

   if (hscroll)
      h -= (GetSystemMetrics(SM_CXHSCROLL)-1);

   int width = 0;
   if (Size)
      {
      for (Element *e = firstControl; e; e = e->next)
         width += e->width;
      if (width++)
         {
         if (width < w)
            {
            // update element positions if needed
            MoveHPos (width-w);
            w = width;
            }
         }
      }

   if (Move || Size)
      {
      UINT flag = SWP_NOZORDER;

      if (!Size)
         flag |= SWP_NOSIZE;
      else if (!Move)
         flag |= SWP_NOMOVE;

      if (width)
         totalWidth = width-w+1;

      // adjust to integral line height
      if (h && lineHeight)
         {
         int old = visibleLines;
         int current = firstLine+activeLine;
         visibleLines = h / lineHeight;
         if (maxLines > 0 && visibleLines > maxLines)
            visibleLines = maxLines;
         if (visibleLines <= 0)
            visibleLines = 1;
         if (visibleLines != old)
            {
            Go (0);
            if (current)
               Go (current);
            }

         if (h != visibleLines * lineHeight + 1)
            h = visibleLines * lineHeight + 1;
         }

      // adjust related controls
      int left = (vbar ? vbar->Attr.W-1 : 0);
      int right = (vscroll ? GetSystemMetrics (SM_CXVSCROLL)-1 : 0);
      int top = (title ? title->Attr.H-1 : 0);
      int bottom = (hscroll ? GetSystemMetrics (SM_CXHSCROLL)-1 : 0);

      if (frame)
         frame->SetWindowPos (HWND_BOTTOM, x-left, y-top,
                              w+left+right-1, h+top+bottom-1, flag);

      if (title)
         title->SetWindowPos (0, x, y-top, w, title->Attr.H, flag);

      if (vbar)
         vbar->SetWindowPos (0, x-left, y, vbar->Attr.W, h, flag);

      flag &= (~SWP_NOMOVE);  // vscroll and hscroll are moved
                              // even when grid does not move
      if (vscroll)
         vscroll->SetWindowPos (0, x+w-1, y, right, h, flag);

      if (hscroll)
         {
         hscroll->SetWindowPos (0, x, y+h-1, w, bottom, flag);
         hscroll->PageMagnitude = w;
         hscroll->SetRange (0, totalWidth-1);
         }
      }
   }


void TGrid::CalcPos ()
   {
   // create the grid's TFont
   SetFont ();
   if (titleLogFont)
      SetTitleFont ();

   // line height
   for (Element *ctrl = firstControl; ctrl; ctrl = ctrl->next)
      {
      int h = ctrl->CtrlHeight ();
      if (h >= lineHeight)
         SetLineHeight (h);
      }
   if (lineHeight == 0)
      lineHeight = 16;  // default height
   visibleLines = Attr.H / lineHeight;

   // column width
   int width = 0;
   int w = 0;
   for (ctrl = firstControl; ctrl; ctrl = ctrl->next)
      {
      w = ctrl->CtrlWidth();
      ctrl->left = width;
      ctrl->width = w+1;

      // make the control a child of TGrid to clip it inside the grid
      TControl *c = ctrl->control;
      c->SetParent (this);
      c->SetWindowPos (0, width, 0, w,
                       ctrl->type == gCombo ? c->Attr.H : lineHeight-1,
                       SWP_NOZORDER|SWP_NOREDRAW);

      width += ctrl->width;
      }

   // resize grid and scroller
   int heightNeeded = lineHeight*visibleLines+1;

   totalWidth = width-Attr.W;
   if (++width > Attr.W)
      width = Attr.W+1;

   if (heightNeeded != Attr.H || width < Attr.W)
      {
      Attr.W = width;
      Attr.H = heightNeeded;
      }

   if (vscroll)
      {
      vscroll->SetWindowPos (HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE|SWP_NOSIZE);
      vscroll->SetRange (0, maxLines < 0 ? nLines-1 : maxLines-1);
      vscroll->PageMagnitude = visibleLines;
      vscroll->SetPosition (0);
      }

   if (title)
      title->SetWindowPos (HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE|SWP_NOSIZE);

   if (vbar)
      vbar->SetWindowPos (HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE|SWP_NOSIZE);

   if (hscroll)
      {
      hscroll->SetWindowPos (HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE|SWP_NOSIZE);
      hscroll->LineMagnitude = LOWORD(GetDialogBaseUnits());// approx. 1 char
      hscroll->SetPosition (0);
      }

   if (frame)
      frame->SetWindowPos (HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE|SWP_NOSIZE);

   // transfer first line of data into the grid, so that controls are
   // properly initialized
   TransferLine (0, tdSetData);

   // SetWindowPos will make sure everything is in place
   SetWindowPos (0, Attr.X, Attr.Y, Attr.W, Attr.H, SWP_NOZORDER);
   }


BOOL TGrid::TransferActiveLine (TTransferDirection td)
	{
	int line = activeLine+firstLine;
	TransferLine (line, td);
   if (td == tdGetData)
	   return IsValidLine (line);
   else
      return TRUE;
	}


BOOL TGrid::CanClose ()
	{
	return TransferActiveLine(tdGetData);  
	}


void TGrid::TransferLine (int line, TTransferDirection dir)
   {
   if (dir == tdGetData)
      {
      for (Element *ctrl = firstControl; ctrl; ctrl = ctrl->next)
         {
         int l = ctrl->GetDataLength();
         char *data = 0;
         if (l > 1)
            {
            data = new char[l];
            ctrl->GetData (data, l);
            }
         TransferItem (line, ctrl->control->GetId (), data);
         if (data)
            delete [] data;
         }
      }

   else if (dir == tdSetData)
      {
      if (maxLines < 0 && line >= nLines)
         {
         if (!AppendNewLine (line))
            return;
         SetNLines (line);
         }

      for (Element *ctrl = firstControl; ctrl; ctrl = ctrl->next)
         {
         char *data = TransferItem (line, ctrl->control->GetId ());
         ctrl->SetData (data);
         }
      }
   }


void TGrid::SetNLines (int n)
   {
   nLines = n;
   if (vscroll)
      vscroll->SetRange (0, nLines-1);
   }


BOOL TGrid::EvSetCursor (HWND hwnd, UINT hitTest, UINT)
   {
   if (hwnd == HWindow && hitTest == HTCLIENT)
      {
      TPoint p;
      GetCursorPos (p);
      ScreenToClient (p);
      int x = p.x+1;

      for (Element *e = firstControl; e; e = e->next)
         {
         int right = e->left+e->width;
         #if 0
         if (x >= right-1 && x <= right+1)
            {
            SetCursor (GridModule, 100);
            return TRUE;
            }
         else
         #endif
         if (x >= e->left && x < right)
            {
            TControl *c = e->control;
            switch (e->type)
               {
               default:
                  SetCursor (0, IDC_ARROW);
                  break;
               case gCombo:
                  if (x < right-15 &&
                      (c->Attr.Style & CBS_DROPDOWNLIST) != CBS_DROPDOWNLIST)
                     SetCursor (0, IDC_IBEAM);
                  else
                     SetCursor (0, IDC_ARROW);
                  break;
               case gEdit:
                  SetCursor (0, IDC_IBEAM);
                  break;
               }
            return TRUE;
            }
         }
      }

   return (BOOL)(DefaultProcessing());
   }


void TGrid::EvLButtonDown (UINT modKeys, TPoint &point)
   {
   ProcessMouse (modKeys, point, WM_LBUTTONDOWN);
   }


void TGrid::EvLButtonDblClk (UINT modKeys, TPoint &point)
   {
   ProcessMouse (modKeys, point, WM_LBUTTONDBLCLK);
   }


void TGrid::ProcessMouse (UINT modKeys, TPoint &point, UINT message)
   {
   Validate ();

   int line = point.y / lineHeight;
   if (line != activeLine)
      Go (firstLine+line);

   Element *e = FindElement (point);
   ClientToScreen (point);
   if (e)
      {
      if (e->type != gStatic)
         {
         TControl *c = e->control;
         if (e != current)
            e->SetFocus ();
         c->ScreenToClient (point);
         c->SendMessage (message, modKeys, MAKELONG(point.x,point.y));
         }
      }
   }


void TGrid::SetFocusOnChild (TControl *child)
   {
   Element *e = FindElement (child);
   if (e == 0 || !e->SetFocus ())
      return;

   int delta = 0;
   if (e->left < 0)
      delta = e->left;
   else if (e->left+e->width > Attr.W)
      delta = e->left+e->width - Attr.W;
   if (delta)
      MoveHPos (delta);
   }


void TGrid::SetHPos (int newpos)
   {
   if (newpos == hpos)
      return;
   MoveHPos (newpos - hpos);
   }


void TGrid::MoveHPos (int delta)
   {
   int pos = hpos + delta;
   if (pos < 0)
      pos = 0;
   else if (pos > totalWidth)
      pos = totalWidth;
   delta = pos - hpos;
   if (delta == 0)
      return;

   for (Element *ctrl = firstControl; ctrl; ctrl = ctrl->next)
      {
      ctrl->left -= delta;
      TControl *c = ctrl->control;
      c->SetWindowPos (0, ctrl->left, c->Attr.Y, 0, 0,
                                       SWP_NOZORDER|SWP_NOSIZE|SWP_NOREDRAW);
      }

   TRect clip (0, 0, Attr.W-2, Attr.H-2);
   ScrollWindowEx (-delta, 0, 0, &clip, 0,0, SW_INVALIDATE);

   if (title)
      {
      clip.Set (0, 0, Attr.W-2, title->Attr.H-2);
      title->ScrollWindowEx (-delta, 0, 0, &clip, 0,0, SW_INVALIDATE);
      title->UpdateWindow ();
      }

   UpdateWindow ();

   if (hscroll)
      hscroll->SetPosition (hpos = pos);
   }


void TGrid::SwapLines (int l1, int l2)
   {
   if (l1 == l2 || l1 < 0 || l2 < 0)
      return;
   if (maxLines >= 0 && (l1 >= maxLines || l2 >= maxLines))
      return;

   TransferLine (firstLine+activeLine, tdGetData);
   for (Element *e = firstControl; e; e = e->next)
      {
      int id = e->control->GetId ();
      string s1 = TransferItem (l1, id);
      string s2 = TransferItem (l2, id);
      TransferItem (l1, id, s2.c_str());
      TransferItem (l2, id, s1.c_str());
      }
   TransferLine (firstLine+activeLine, tdSetData);

   int line = l1 - firstLine;
   if (line >= 0 && line < visibleLines)
      {
      int y = line*lineHeight;
      InvalidateRect (TRect (0, y, Attr.W-2, y+lineHeight-1), FALSE);
      }

   line = l2 - firstLine;
   if (line >= 0 && line < visibleLines)
      {
      int y = line*lineHeight;
      InvalidateRect (TRect (0, y, Attr.W-2, y+lineHeight-1), FALSE);
      }
   }


void TGrid::RefreshLine (int line)
   {
   if (line < firstLine || line >= firstLine+visibleLines)
      return;

   int gridline = line - firstLine;
   int y = lineHeight * gridline;
   TRect clip (0, y, Attr.W-2, y+lineHeight);
   InvalidateRect (clip, FALSE);

   if (gridline == activeLine)
      TransferLine (line, tdSetData);

   // update vbar
   RefreshItem (line, 0);
   }


void TGrid::RefreshItem (int line, int id)
   {
   if (line < firstLine || line >= firstLine+visibleLines)
      return;

   int gridline = line - firstLine;

   if (gridline == activeLine)
      {
      Element *e = FindElement (id);
      if (e)
         e->SetData (TransferItem (line, id));
      }

   int y = lineHeight * gridline;
   if (id != 0)   // update grid data
      {
      Element *e = FindElement (id);
      TRect clip (e->left, y, e->left+e->width, y+lineHeight);
      if (e->type == gCombo)  // don't redraw down-arrow
         clip.right -= ((LOWORD(GetDialogBaseUnits())) << 1);
      if (clip.right < 0 || clip.left >= Attr.W)
         return;
      if (clip.left < 0)
         clip.left = 0;
      if (clip.right >= Attr.W-1)
         clip.right = Attr.W - 2;
      if (!clip.IsEmpty ())
         InvalidateRect (clip, FALSE);
      }
   else if (vbar)  // update vbar
      {
      TRect clip (0, y, vbar->Attr.W-2, y+lineHeight);
      vbar->InvalidateRect (clip, FALSE);
      }
   }


void TGrid::Refresh () 
   {
   for (int i = 0; i < visibleLines; i++)
      RefreshLine (firstLine+i);
   }


void TGrid::ChangeTitle (int id, const char *newTitle)
   {
   Element *e = FindElement (id);
   if (e)
      e->SetTitle (newTitle);
   }

   
const char *TGrid::GetTitle (int id) const
   {
   Element *e = FindElement (id);
   if (e)
      return e->title.c_str();
   else
      return 0;
   }


void TGrid::SetLineMarker (TGrid::LineMarker lm)
   {
   if (lm == lineMarker)
      return;

   // remove old markers
   if (vbar)
      vbar->ShowText (activeLine);

   lineMarker = lm;

   if (vbar)
      vbar->ShowMarker (activeLine);
   }


void TGrid::SetVBarText (int cwidth)
   {
   if (vbar)
      vbar->SetUserText (cwidth);
   }


void TGrid::SetFont (char *facename, int height, int weight, BOOL italic)
   {
   if (!IsWindow())
      {
      strcpy (gridLogFont->lfFaceName, facename);
      if (height)
         gridLogFont->lfHeight = height;
      gridLogFont->lfWeight = weight;
      gridLogFont->lfItalic = BYTE(italic);
      }
   }


void TGrid::SetTitleFont (char *facename, int height, int weight, BOOL italic)
   {
   if (!IsWindow ())
      {
      if (titleLogFont == 0)
         {
         titleLogFont = new LOGFONT;
         memcpy (titleLogFont, gridLogFont, sizeof(LOGFONT));
         }
      strcpy (titleLogFont->lfFaceName, facename);
      if (height)
         titleLogFont->lfHeight = height;
      titleLogFont->lfWeight = weight;
      titleLogFont->lfItalic = BYTE(italic);
      }
   }


void TGrid::SetFont ()
   {
   if (gridFont)
      delete gridFont;
   gridFont = new TFont (gridLogFont);

   // change fonts and redraw elements
   WPARAM wp = reinterpret_cast<WPARAM>(HFONT(*gridFont));
   LPARAM lp = MAKELPARAM (TRUE, 0);
   SendMessage (WM_SETFONT, wp, lp);
   for (Element *ctrl = firstControl; ctrl; ctrl = ctrl->next)
      if (ctrl->control && ctrl->control->HWindow)
         ctrl->control->SendMessage (WM_SETFONT, wp, lp);
   }


void TGrid::SetTitleFont ()
   {
   if (titleFont)
      delete titleFont;
   if (titleLogFont == 0)
      {
      titleLogFont = new LOGFONT;
      memcpy (titleLogFont, gridLogFont, sizeof(LOGFONT));
      }
   titleFont = new TFont (titleLogFont);

   WPARAM wp = reinterpret_cast<WPARAM>(HFONT(*titleFont));
   LPARAM lp = MAKELPARAM (TRUE, 0);
   SendMessage (WM_SETFONT, wp, lp);
   if (vbar)
      vbar->SendMessage (WM_SETFONT, wp, lp);
   }

// EOF
